//@ts-check
"use strict";

/**
 * @typedef QuestName
 * @type {keyof typeof questDescs}
 */

/**
 * @typedef CounterCond
 * @type {{
 *   prop: keyof Stats;
 *   min: number;
 * }}
 */

/**
 * @typedef QuestConditions
 * @type {{
 *   anchor: NPCName,
 *   quests?: QuestName[],
 *   nearbyNPCs?: NPCName[],
 *   items?: ItemName[],
 *   counters?: CounterCond[],
 *   bools?: (keyof Stats)[],
 *   minFollowers?: number;
 * }}
 */

/**
 * @typedef QuestDesc
 * @type {{
 *   id: QuestName,
 *   conditions: QuestConditions,
 *   outcome: {
 *     items?: ItemName[],
 *     followers?: NPCName[],
 *     bools?: (keyof Stats)[],
 *   }
 * }}
 */

/**
 * @type {{
 *   grandmaMinion: QuestDesc,
 *   gudulDogRescue: QuestDesc,
 *   sheepSlaughter: QuestDesc,
 *   conanFights: QuestDesc,
 *   samuraiCoward: QuestDesc,
 *   greedyBro: QuestDesc,
 *   shipRepair: QuestDesc,
 *   concert: QuestDesc,
 *   saveMom: QuestDesc,
 * }}
 */
const questDescs = {
  grandmaMinion: {
    id: "grandmaMinion",
    conditions: {
      anchor: "grandma",
      nearbyNPCs: ["onion"],
    },
    outcome: {
      followers: ["grandma"],
    },
  },
  gudulDogRescue: {
    id: "gudulDogRescue",
    conditions: {
      anchor: "gudul",
      nearbyNPCs: ["pup"],
    },
    outcome: {
      followers: ["gudul"],
    },
  },
  sheepSlaughter: {
    id: "sheepSlaughter",
    conditions: {
      anchor: "farmhand",
      counters: [{ prop: "sheepKilled", min: 12 }]
    },
    outcome: {}, // TODO
  },
  conanFights: {
    id: "conanFights",
    conditions: {
      anchor: "conan",
      bools: ["killedAllBaddiesOnce"],
    },
    outcome: {
      items: ["torch"],
    }
  },
  samuraiCoward: {
    id: "samuraiCoward",
    conditions: {
      anchor: "conan",
      items: ["torch"],
    },
    outcome: {
      followers: ["samurai"],
    }
  },
  greedyBro: {
    id: "greedyBro",
    conditions: {
      anchor: "broody",
      items: ["money"],
    },
    outcome: {
      followers: ["broody"]
    }
  },
  shipRepair: {
    id: "shipRepair",
    conditions: {
      anchor: "captain",
      items: ["boardBag"],
    },
    outcome: {
      bools: ["shipRepaired"]
    },
  },
  concert: {
    id: "concert",
    conditions: {
      anchor: "bard",
      minFollowers: 2,
    },
    outcome: {
      followers: ["bard"],
    }
  },
  saveMom: {
    id: "saveMom",
    conditions: {
      anchor: "bro",
      nearbyNPCs: ["mom"],
    },
    outcome: {
      followers: ["mom"],
    },
  }
}


/**
 * @param {QuestConditions} conds
 * @returns {QuestConditions}
 */
function filterDoneQuestConditions(conds) {
  const { stats } = inGame;

  /** @type {QuestConditions} */
  let res = {
    anchor: conds.anchor,
    counters: [],
    bools: [],
    items: [],
    nearbyNPCs: [],
    quests: [],
  };

  if (conds.quests) {
    for (const quest of conds.quests) {
      if (stats.questsCompleted[quest]) {
        // already completed prerequisite quest
      } else {
        res.quests.push(quest);
      }
    }
  }

  if (conds.items) {
    for (const item of conds.items) {
      if (stats.itemSlots.findIndex((slot) => slot.item === item) !== -1) {
        // got item
      } else {
        res.items.push(item);
      }
    }
  }

  if (conds.nearbyNPCs) {
    for (const npc of conds.nearbyNPCs) {
      let anchor = findNPC(conds.anchor);
      if (anchor) {
        if (inGame.activeWorld.npcs.findIndex((worldNpc) => {
          if (worldNpc.desc.id !== npc) {
            return false;
          }
          if (worldNpc.following) {
            return true;
          }
          let dist = distxy(anchor.pos, worldNpc.pos);
          if (dist <= 4) {
            console.log(`succeeding nearbyNPC, dist = ${dist}`);
            return true;
          } else {
            console.log(`failing nearbyNPC, dist = ${dist}`);
          }
          return false;
        }) !== -1) {
          // the right npc is nearby
        } else {
          res.nearbyNPCs.push(npc);
        }
      } else {
        // anchor is not in active world, can't check nearby NPCs
        res.nearbyNPCs = [...conds.nearbyNPCs];
      }
    }
  }

  if (conds.counters) {
    for (const counter of conds.counters) {
      let value = stats[counter.prop];
      if (value >= counter.min) {
        // satisfied
      } else {
        res.counters.push(counter);
      }
    }
  }

  if (conds.bools) {
    for (const bool of conds.bools) {
      let value = stats[bool];
      if (value === true) {
        // satisfied
      } else {
        res.bools.push(bool);
      }
    }
  }

  if (typeof conds.minFollowers === "number") {
    if (countFollowers() > conds.minFollowers) {
      // satisfied
    } else {
      res.minFollowers = conds.minFollowers;
    }
  }

  return res;
}

/**
 * @param {QuestName} questName
 * @returns {boolean}
 */
function isQuestDone(questName) {
  if (wasQuestAlreadyDone(questName)) {
    return true;
  }

  let desc = questDescs[questName];

  let filteredConds = filterDoneQuestConditions(desc.conditions);
  console.table({ conds: desc.conditions, filtered: filteredConds });
  return (
    isEmpty(filteredConds.counters) &&
    isEmpty(filteredConds.bools) &&
    isEmpty(filteredConds.nearbyNPCs) &&
    isEmpty(filteredConds.items) &&
    isEmpty(filteredConds.quests) &&
    (!filteredConds.minFollowers)
  );
}

/**
 * @param {QuestName} questName
 * @returns {boolean}
 */
function wasQuestAlreadyDone(questName) {
  return inGame.stats.questsCompleted[questName];
}

/**
 * @param {QuestName} questName
 */
function applyQuestOutcome(questName) {
  const { stats } = inGame;
  if (stats.questsCompleted[questName]) {
    return;
  }
  notify("huge", getText(`questMeta.questCompleted`) + getText(`quests.${questName}.name`));
  getSound("quests.complete").play();

  let desc = questDescs[questName];

  stats.questsCompleted[desc.id] = true;
  delete stats.questsInProgress[desc.id];

  for (const item of (desc.conditions.items || [])) {
    dropItem(item);
  }

  for (const item of (desc.outcome.items || [])) {
    pickupItem(item);
  }
  for (const npc of (desc.outcome.followers || [])) {
    stats.followers[npc] = true;
    findNPC(npc).following = true;
  }
  for (const bool of (desc.outcome.bools || [])) {
    // @ts-ignore
    stats[bool] = true;
  }
}

/**
 * @param {any[] | null | undefined} arr
 * @returns {boolean}
 */
function isEmpty(arr) {
  if (arr === null || typeof arr === "undefined") {
    return true;
  }
  if (arr.length == 0) {
    return true;
  }
  return false;
}

/**
 * @returns {number}
 */
function countFollowers() {
  let count = 0;
  const { stats } = inGame;
  for (const k of Object.keys(stats.followers)) {
    if (stats.followers[k] === true) {
      count++;
    }
  }
  return count;
}

const MAX_FOLLOWERS = 5;

/**
 * @param {NPC} npc
 * @returns {boolean}
 */
function addFollower(npc) {
  if (countFollowers() >= MAX_FOLLOWERS) {
    return false;
  }

  npc.following = true;
  inGame.stats.followers[npc.desc.id] = true;
  return true;
}

/**
 * @param {QuestName} questName
 * @returns {boolean}
 */
function startQuest(questName) {
  if (isQuestDone(questName)) {
    console.log(`Not starting quest, done already: ${questName}`);
    return false;
  }

  if (isQuestInProgress(questName)) {
    console.log(`Not starting quest, already in progress: ${questName}`);
    return false;
  }

  inGame.stats.questsInProgress[questName] = true;
  notify("huge", getText(`questMeta.questStarted`) + getText(`quests.${questName}.name`));
  getSound("quests.new").play();

  return true;
}

/**
 * @param {QuestName} questName
 * @returns {boolean}
 */
function isQuestInProgress(questName) {
  return !!inGame.stats.questsInProgress[questName];
}

/**
 * @param {QuestName} questName
 * @returns {boolean}
 */
function justCompletedQuest(questName) {
  if (!wasQuestAlreadyDone(questName) && isQuestDone(questName)) {
    applyQuestOutcome(questName);
    return true;
  }

  return false;
}

/**
 * @param {QuestName} questName
 * @returns {boolean}
 */
function canSuggestQuest(questName) {
  if (isQuestDone(questName) || isQuestInProgress(questName)) {
    return false;
  }

  return true;
}
